/*
** glconfig.c 
**
** Written by Ian Buck
** 6/15/2000
** http://graphics.stanford.edu/software/wiregl
**
** Copyright 2000 
** The Board of Trustees of The Leland Stanford Junior University.
** All rights reserved.
**
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or
** modify this software, provided that the above copyright notice and
** this permission notice appear in all copies of this software.  No
** part of this software or any derivatives thereof may be used in
** graphics systems for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind,
** express, implied or otherwise, including without limitation, any
** warranty of merchantability or fitness for a particular purpose.
*/

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>

#include "gltypes.h"
#include "glconfig.h"
#include "glcontext.h"
#include "glhw.h"

#define GLCONFIG_VENDOR "GLConfig 1.0 by Ian Buck"
#define GLCONFIG_RENDERER "GLConfig"
#define GLCONFIG_VERSION "1.1"
#define GLCONFIG_EXTENSIONS ""

void
__glconfig_flushnop (GLcontext *c) {
	UNUSED(c);
}

void 
__glconfig_vpinchnop (GLcontext *c) {
	UNUSED(c);
}

GLconfig *
__glconfig_CreateDefaultConfig (void){
	GLconfig *c = (GLconfig *) malloc (sizeof (GLconfig));
	memset((void *) c, 0, sizeof (GLconfig));
	
	c->vendor = strdup (GLCONFIG_VENDOR);
	c->renderer = strdup (GLCONFIG_RENDERER);
	c->version = strdup (GLCONFIG_VERSION);
	c->extensions = strdup (GLCONFIG_EXTENSIONS);

	c->maxlight = 8;
	c->max3dtexturesize = 16;
	c->maxnamestackdepth = 64;
	c->maxlistnesting = 64;
	c->maxevalorder = 8;
	c->maxattribstackdepth = 16;
	c->maxclientattribstackdepth = 16;
	c->auxbuffers = 0;
	c->rgbamode = GL_TRUE;
	c->indexmode = GL_FALSE;
	c->doublebuffer = GL_TRUE;
	c->stereo = GL_FALSE;
	c->subpixelbits = 4;
	c->redbits = 8;
	c->greenbits = 8;
	c->bluebits = 8;
	c->alphabits = 8;
	c->depthbits = 24;
	c->stencilbits = 32;
	c->accumredbits = 32;
	c->accumgreenbits = 32;
	c->accumbluebits = 32;
	c->accumalphabits = 32;
	c->aliasedpointsizerange_min = 1.0f;
	c->aliasedpointsizerange_max = 1.0f;
	c->aliasedpointsizegranularity = 1.0f;
	c->smoothpointsizerange_min = 1.0f;
	c->smoothpointsizerange_max = 1.0f;
	c->smoothpointgranularity = 1.0f;
	c->aliasedlinewidth_min = 1.0f;
	c->aliasedlinewidth_max = 1.0f;
	c->aliasedlinegranularity = 1.0f;
	c->smoothlinewidth_min = 1.0f;
	c->smoothlinewidth_max = 1.0f;
	c->smoothlinegranularity = 1.0f;
	c->maxconvolution_width = 3;
	c->maxconvolution_height = 3;
	c->maxelementsindices = GL_MAXINT;
	c->maxelementsvertices = GL_MAXINT;
	c->maxpixelmaptable = 32;
	c->maxmodelviewstackdepth = 64;
	c->maxprojectionstackdepth = 8;
	c->maxtexturestackdepth = 8;
	c->maxcolorstackdepth = 8;
	c->maxclipplanes = 6;
	c->maxviewportdims_width = 2048;
	c->maxviewportdims_height = 2048;
	c->maxtexturesize = 512;

	c->numprojectors = 1;
	c->bounds[0].x1 = 0;
	c->bounds[0].y1 = 0;
	c->bounds[0].x2 = 0;
	c->bounds[0].y2 = 0;
	c->id = 0;
	c->mode = GLCONFIG_STATE;

	c->viewport.x1 = 0;
	c->viewport.y1 = 0;
	c->viewport.x2 = 0;
	c->viewport.y2 = 0;
	c->localviewport = c->viewport;

	c->flush = __glconfig_flushnop;
	c->vpinch = __glconfig_vpinchnop;
	c->updateclientpointer = NULL;
	c->immediateapi = NULL;
	c->packapi = NULL;

	c->vertexBBapi = NULL;
	c->vertexCapi = NULL;

	c->current = NULL;

	c->gui_error = (getenv("WIREGL_GUI_ERROR")!=NULL);
	c->draw_bbox = (getenv("WIREGL_DRAW_BBOX")!=NULL);
	c->bucket_size = GL_MAXINT;
	c->broadcast = (getenv("WIREGL_BROADCAST")!=NULL);
	c->pack_only = (getenv("WIREGL_PACK_ONLY")!=NULL);
	c->resolve_arrays = (getenv("WIREGL_RESOLVE_ARRAYS")!=NULL);
	c->state_error_gui = (getenv("WIREGL_STATE_ERROR_GUI")!=NULL);

	return c;
}

GLconfig *
__glconfig_CopyConfig (GLconfig *a) {
	GLconfig *c;
	GLconfigmsg *msg;

	msg = __glconfig_CreateMessage (a);
	c = __glconfig_CreateMessageConfig (msg);
	__glconfig_FreeMessage(msg);

	return c;
}

GLconfig *
__glconfig_CreateSystemConfig(void) {
	GLconfig *c;
	GLint irange[2];
	GLfloat frange[2];
	const char *vendor;
	const char *renderer;
	const char *version;
	const char *extensions;

	c = __glconfig_CreateDefaultConfig ();
	free (c->vendor);
	free (c->renderer);
	free (c->version);
	free (c->extensions);

	// Make sure that __glhw functions are valid.
	__glhw_LoadSwitchApi(NULL);

	__glhw_GetIntegerv(GL_MAX_LIGHTS, &(c->maxlight));
	__glhw_GetIntegerv(GL_MAX_CLIP_PLANES, &(c->maxclipplanes));
	__glhw_GetIntegerv(GL_MAX_COLOR_STACK_DEPTH, &(c->maxcolorstackdepth));
	__glhw_GetIntegerv(GL_MAX_MODELVIEW_STACK_DEPTH, &(c->maxmodelviewstackdepth));
	__glhw_GetIntegerv(GL_MAX_PROJECTION_STACK_DEPTH, &(c->maxprojectionstackdepth));
	__glhw_GetIntegerv(GL_MAX_TEXTURE_STACK_DEPTH, &(c->maxtexturestackdepth));
	__glhw_GetIntegerv(GL_SUBPIXEL_BITS, &(c->subpixelbits));
	__glhw_GetIntegerv(GL_MAX_3D_TEXTURE_SIZE, &(c->max3dtexturesize));
	__glhw_GetIntegerv(GL_MAX_TEXTURE_SIZE, &(c->maxtexturesize));
	__glhw_GetIntegerv(GL_MAX_PIXEL_MAP_TABLE, &(c->maxpixelmaptable));
	__glhw_GetIntegerv(GL_MAX_NAME_STACK_DEPTH, &(c->maxnamestackdepth));
	__glhw_GetIntegerv(GL_MAX_LIST_NESTING, &(c->maxlistnesting));
	__glhw_GetIntegerv(GL_MAX_EVAL_ORDER, &(c->maxevalorder));
	__glhw_GetIntegerv(GL_MAX_VIEWPORT_DIMS, irange);
	c->maxviewportdims_width = irange[0];
	c->maxviewportdims_height = irange[1];
	__glhw_GetIntegerv(GL_MAX_ATTRIB_STACK_DEPTH, &(c->maxattribstackdepth));
	__glhw_GetIntegerv(GL_MAX_CLIENT_ATTRIB_STACK_DEPTH, &(c->maxclientattribstackdepth));
	__glhw_GetIntegerv(GL_AUX_BUFFERS, &(c->auxbuffers));
	__glhw_GetBooleanv(GL_RGBA_MODE, &(c->rgbamode));
	__glhw_GetBooleanv(GL_INDEX_MODE, &(c->indexmode));
	__glhw_GetBooleanv(GL_DOUBLEBUFFER, &(c->doublebuffer));
	__glhw_GetBooleanv(GL_STEREO, &(c->stereo));
/* Arg...  1.2 spec has two separte ones 
	
	__glhw_GetFloatv(GL_ALIASED_POINT_SIZE_RANGE, frange);
	c->aliasedlinewidth_min = frange[0];
	c->aliasedlinewidth_max = frange[1];
	__glhw_GetFloatv(GL_ALIASED_POINT_GRANULARITY, &(c->aliasedlinegranularity));
	__glhw_GetFloatv(GL_SMOOTH_POINT_SIZE_RANGE, frange);
	c->smoothlinewidth_min = frange[0];
	c->smoothlinewidth_max = frange[1];
	__glhw_GetFloatv(GL_SMOOTH_POINT_GRANULARITY, &(c->smoothlinegranularity));

	*/

	__glhw_GetFloatv(GL_POINT_SIZE_RANGE, frange);
	c->aliasedlinewidth_min = frange[0];
	c->aliasedlinewidth_max = frange[1];
	c->smoothlinewidth_min = frange[0];
	c->smoothlinewidth_max = frange[1];

	//__glhw_GetConvolutionParameteriv(GL_CONVOLUTION_2D, GL_MAX_CONVOLUTION_WIDTH, &(c->maxconvolution_width));
	//__glhw_GetConvolutionParameteriv(GL_CONVOLUTION_2D, GL_MAX_CONVOLUTION_HEIGHT, &(c->maxconvolution_height));
	__glhw_GetIntegerv(GL_MAX_ELEMENTS_INDICES, &(c->maxelementsindices));
	__glhw_GetIntegerv(GL_MAX_ELEMENTS_VERTICES, &(c->maxelementsvertices));

	__glhw_GetIntegerv(GL_RED_BITS, &(c->redbits));
	__glhw_GetIntegerv(GL_GREEN_BITS, &(c->greenbits));
	__glhw_GetIntegerv(GL_BLUE_BITS, &(c->bluebits));
	__glhw_GetIntegerv(GL_ALPHA_BITS, &(c->alphabits));
	__glhw_GetIntegerv(GL_DEPTH_BITS, &(c->depthbits));
	__glhw_GetIntegerv(GL_STENCIL_BITS, &(c->stencilbits));
	__glhw_GetIntegerv(GL_ACCUM_RED_BITS, (&c->accumredbits));
	__glhw_GetIntegerv(GL_ACCUM_GREEN_BITS, (&c->accumgreenbits));
	__glhw_GetIntegerv(GL_ACCUM_BLUE_BITS, (&c->accumbluebits));
	__glhw_GetIntegerv(GL_ACCUM_ALPHA_BITS, (&c->accumalphabits));

	vendor = (const char *) __glhw_GetString(GL_VENDOR);
	renderer = (const char *) __glhw_GetString(GL_RENDERER);
	version = (const char *) __glhw_GetString(GL_VERSION);
	extensions = (const char *) __glhw_GetString(GL_EXTENSIONS);
	c->vendor = strdup (vendor);
	c->renderer = strdup (renderer);
	c->version = strdup (version);
	c->extensions = strdup (extensions);

	return c;
}

void
__glconfig_FreeConfig (GLconfig * c) {
#if 0
	/* XXX: Should free here but causes
	** problems with custom user 
	** malloced strings 
	*/
	if (c->vendor)
		free (c->vendor);
	if (c->renderer)
		free (c->renderer);
	if (c->version)
		free (c->version);
	if (c->extensions)
		free (c->extensions);
#endif
	free (c);
}

GLconfigmsg * 
__glconfig_CreateMessage (GLconfig *c) {
	GLconfigmsg *msg;
	int size;
	char *p;
	size =  sizeof (GLconfigmsg) +
			sizeof (GLconfig) + 
			strlen(c->vendor) + 1 +
			strlen(c->renderer) + 1 +
			strlen(c->version) + 1 +
			strlen(c->extensions) + 1;

	msg = (GLconfigmsg *) malloc (size);
	msg->size = size;

	memcpy ((void *)msg->data, (const void *) c, sizeof (GLconfig));
	p = msg->data + sizeof (GLconfig);
	strcpy (p, c->vendor);
	p+=strlen(c->vendor) + 1;
	strcpy (p, c->renderer);
	p+=strlen(c->renderer) + 1;
	strcpy (p, c->version);
	p+=strlen(c->version) + 1;
	strcpy (p, c->extensions);
	return msg;
}

GLconfig * 
__glconfig_CreateMessageConfig(GLconfigmsg *m) {
	
	GLconfig *c;
	char *vendor;
	char *renderer;
	char *version;
	char *extensions;
	vendor = m->data + sizeof (GLconfig);
	renderer = vendor + strlen(vendor) + 1;
	version = renderer + strlen (renderer) + 1;
	extensions = version + strlen (version) + 1;

	c = __glconfig_CreateDefaultConfig ();
	free (c->vendor);
	free (c->renderer);
	free (c->version);
	free (c->extensions);

	memcpy ((void *) c, (const void *) m->data, sizeof (GLconfig));

#if 0
	c->flush = __glconfig_flushnop;
	c->immediateapi = NULL;
	c->packapi = NULL;
	c->stateapi = NULL;
#endif

	c->vendor = (char *) malloc (strlen (vendor) + 1);
	c->renderer = (char *) malloc (strlen (renderer) + 1);
	c->version = (char *) malloc (strlen (version) + 1);
	c->extensions = (char *) malloc (strlen (extensions) + 1);

	strcpy (c->vendor, vendor);
	strcpy (c->renderer, renderer);
	strcpy (c->version, version);
	strcpy (c->extensions, extensions);

	return c;
}

void 
__glconfig_FreeMessage(GLconfigmsg *m) {
	free (m);
}
