/*
** glviewport.c
**
** Written by Ian Buck
** 6/15/2000
** http://graphics.stanford.edu/software/wiregl
**
** Copyright 2000 
** The Board of Trustees of The Leland Stanford Junior University.
** All rights reserved.
**
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or
** modify this software, provided that the above copyright notice and
** this permission notice appear in all copies of this software.  No
** part of this software or any derivatives thereof may be used in
** graphics systems for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind,
** express, implied or otherwise, including without limitation, any
** warranty of merchantability or fitness for a particular purpose.
*/

#include "glcontext.h"
#include "glerror.h"
#include "glhw.h"

void 
__glviewport_initbits (GLviewportbits *v, GLconfig *cfg) {
	UNUSED(v);
	UNUSED(cfg);
	/* do nothing */
}

void
__glviewport_clamp (GLviewportstate *current, GLviewportstate *target, GLrecti *p) {
	p->x1 = target->x;
	p->y1 = target->y;
	p->x2 = target->x + target->w;
	p->y2 = target->y + target->h;

	if (current->outputdims.x1 == current->outputdims.x2 &&
		current->outputdims.y1 == current->outputdims.y2)
		return;

	/* Clamp it to the outputdims */
	if (p->x1 < current->outputdims.x1) p->x1 = current->outputdims.x1;
	if (p->y1 < current->outputdims.y1) p->y1 = current->outputdims.y1;
	if (p->x1 > current->outputdims.x2) p->x1 = current->outputdims.x2;
	if (p->y1 > current->outputdims.y2) p->y1 = current->outputdims.y2;

	if (p->x2 < current->outputdims.x1) p->x2 = current->outputdims.x1;
	if (p->y2 < current->outputdims.y1) p->y2 = current->outputdims.y1;
	if (p->x2 > current->outputdims.x2) p->x2 = current->outputdims.x2;
	if (p->y2 > current->outputdims.y2) p->y2 = current->outputdims.y2;
	
	if (p->x2 == p->x1 ||
		p->y2 == p->y1) {
		p->x1 = 0;
		p->x2 = 0;
		p->y1 = 0;
		p->y2 = 0;
	} else {
		p->x1 -= current->outputdims.x1;
		p->y1 -= current->outputdims.y1;
		p->x2 -= current->outputdims.x1;
		p->y2 -= current->outputdims.y1;
	}
}

void
__glviewport_init(GLviewportstate *v, GLconfig *c) {
	int id;	
	v->maxviewportdims_width = c->maxviewportdims_width;
	v->maxviewportdims_height = c->maxviewportdims_height;

	v->x = c->viewport.x1;
	v->y = c->viewport.y1;
	v->w = c->viewport.x2 - c->viewport.x1;
	v->h = c->viewport.y2 - c->viewport.y1;

	id = c->id;
	if (id >= 0 && c->numprojectors > 0)
		v->outputdims = c->bounds[id];
	else
		v->outputdims = c->viewport;

	if (!c->localviewport.x2) 
		v->widthscale = 1.0f;
	else
		v->widthscale = ((GLfloat) v->w) / ((GLfloat)c->localviewport.x2);

	if (!c->localviewport.y2)
		v->heightscale = 1.0f;
	else
		v->heightscale = ((GLfloat) v->h) / ((GLfloat)c->localviewport.y2);

	v->f = 1.0;
	v->n = 0.0;
}

void
__glviewport_destroy(GLviewportstate *v) {
/*
** Do nothing...
*/
	UNUSED(v);
}

void
__glviewport_ApplyViewport(GLviewportstate *v, GLvectorf *p) {
	p->x = (p->x+1.0f)*(v->w / 2.0f) + v->x;
	p->y = (p->y+1.0f)*(v->h / 2.0f) + v->y;
}

void GLSTATE_DECL
__glstate_Viewport(GLint x, GLint y, GLsizei width, GLsizei height) {
	GLcontext *g = GetCurrentContext();
	GLviewportstate *v = &(g->viewport);
	GLstatebits *sb = GetStateBits();
	GLviewportbits *vb = &(sb->viewport);
	GLtransbits *tb = &(sb->trans);
	
	if (g->current.beginend)
		if (__glerror(__LINE__, __FILE__, GL_INVALID_OPERATION, "glViewport called in Begin/End"))
			return;
	
	if (IsBucketingActive()) 
		g->bucket.flush(g);

	if (width < 0 || height < 0)
		if (__glerror(__LINE__, __FILE__, GL_INVALID_VALUE, "Negative viewport width or height: %dx%d", width, height))
			return;

	if (x > v->maxviewportdims_width)	x = v->maxviewportdims_width;
	if (x < -v->maxviewportdims_width)	x = -v->maxviewportdims_width;
	if (y > v->maxviewportdims_height)	y = v->maxviewportdims_height;
	if (y < -v->maxviewportdims_height)	y = -v->maxviewportdims_height;
	if (width > v->maxviewportdims_width)	width = v->maxviewportdims_width;
	if (height > v->maxviewportdims_height)	height = v->maxviewportdims_height;

	v->x = (GLint) (x*v->widthscale+0.5f);
	v->y = (GLint) (y*v->heightscale+0.5f);
	v->w = (GLint) (width*v->widthscale+0.5f);
	v->h = (GLint) (height*v->heightscale+0.5f);

	__glbucket_bounds(v->x, v->y, v->w, v->h);

	vb->dims = g->nbitID;
	vb->dirty = g->nbitID;
	tb->base = g->nbitID;
	tb->compress = g->nbitID;
	tb->dirty = g->nbitID;

}

void GLSTATE_DECL
__glstate_DepthRange(GLclampd znear, GLclampd zfar) {
	GLcontext *g = GetCurrentContext();
	GLviewportstate *v = &(g->viewport);
	GLstatebits *sb = GetStateBits();
	GLviewportbits *vb = &(sb->viewport);
	GLtransbits *tb = &(sb->trans);

	if (g->current.beginend)
		if (__glerror(__LINE__, __FILE__, GL_INVALID_OPERATION, "glDepthRange called in Begin/End"))
			return;

	if (IsBucketingActive()) 
		g->bucket.flush(g);

	v->n = znear;
	v->f = zfar;
	if (v->n < 0.0) v->n = 0.0;
	if (v->n > 1.0) v->n = 1.0;
	if (v->f < 0.0) v->f = 0.0;
	if (v->f > 1.0) v->f = 1.0;

	vb->depth = g->nbitID;
	vb->dirty = g->nbitID;
	tb->compress = g->nbitID;
	tb->dirty = g->nbitID;
}

void GLSTATE_DECL
__gltrack_Viewport(GLint x, GLint y, GLsizei width, GLsizei height) {
	GLcontext *g = GetCurrentContext();
	GLviewportstate *v = &(g->viewport);
	GLstatebits *sb = GetStateBits();
	GLviewportbits *vb = &(sb->viewport);

	__glhw_Viewport(x, y, width, height);

	v->x = x;
	v->y = y;
	v->w = width;
	v->h = height;

	vb->dims = g->nbitID;
	vb->dirty = g->nbitID;
}

void GLSTATE_DECL
__gltrack_DepthRange(GLclampd znear, GLclampd zfar) {
	GLcontext *g = GetCurrentContext();
	GLviewportstate *v = &(g->viewport);
	GLstatebits *sb = GetStateBits();
	GLviewportbits *vb = &(sb->viewport);

	__glhw_DepthRange( znear,  zfar);

	v->n = znear;
	v->f = zfar;

	vb->depth = g->nbitID;
	vb->dirty = g->nbitID;
}





