/* 
** include/wiregl_client.h 
** http://graphics.stanford.edu/software/wiregl
**
** Copyright 2000
** The Board of Trustees of The Leland Stanford Junior University.
** All rights reserved.
**
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or
** modify this software, provided that the above copyright notice and
** this permission notice appear in all copies of this software.  No
** part of this software or any derivatives thereof may be used in
** graphics systems for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind,
** express, implied or otherwise, including without limitation, any
** warranty of merchantability or fitness for a particular purpose.
*/

#ifndef WIREGL_CLIENT_H
#define WIREGL_CLIENT_H

#ifdef WINDOWS
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include "wiregl/include/wiregl_util.h"
#include "wiregl/glcontext/glcontext.h"
#include "wiregl/glcontext/glhw.h"
#include "wiregl/include/wiregl_bbox.h"
#include "wiregl/include/wiregl_opcodes.h"
#include "wiregl/include/wiregl_feather.h"

typedef struct WireGLSendBuffer {
	void          *pack;
	unsigned int   size;
	unsigned char *data_start, *data_current, *data_end;
	unsigned char *opcode_start, *opcode_current, *opcode_end;
} WireGLSendBuffer;

typedef struct WireGLPipeServer {
	WireGLConnection *conn;
	int w,h;
	int mural_x, mural_y;
	int window_x, window_y;
	WireGLFeatherParam feather;
	GLcontext *context;

	char *server;
	int index;

	unsigned int writeback;

	int use_l2;
	int l2_width, l2_height;

	WireGLSendBuffer buffer;
} WireGLPipeServer;

typedef struct WireGLGlobals {
	/* buffer state */
	WireGLSendBuffer buffer;
	GLcontext       *context;

#ifdef SIMD
	__m128 pad16byte;
#endif
	GLvectorf bounds_min;
	GLvectorf bounds_max;
	/* tracking state */
	GLcurrent current;

	/* pipe server state */
	int num_servers;
	WireGLPipeServer *servers;

	/* client state */
#ifdef WINDOWS
	HDC  client_hdc;
	HWND client_hwnd;
#else
	void         *glx_display;  /* (Display *) */
	unsigned int  glx_drawable; /* (GLXDrawable -> XID -> CARD32) */
#endif

	/* instrumentation */
	struct {
		int bytes_per_server;
		int bytes_per_frame_per_server;
		int min_bytes_sent;
		int frame_rate;
		int flush_time;
		int send_time;
		int diff_context_time;
		int total_bytes_sent;
		int total_time;
	} print;

	/* miscellany */
	struct {
		int width;
		int height;
	} mural;

	int pipe_buffer_size;
	int depth_bits;
	int stencil_bits;	
	int normal_compression;
	int vertex_compression;
	int unsafe_compression;
	int compression_depthbits;
	int swap_barrier;
	char *barrier_server;
	int gui_error;
	int draw_bbox;
	char *warn_level_str;
	int bucket_size;
	int broadcast;
	int pack_only;
	int resolve_arrays;
	int state_error_gui;
	int sync_on_finish;
	int sync_on_swap;
	int split_begin_end;
	int use_ring;
	int use_simd;

} WireGLGlobals;

extern WireGLGlobals __wiregl_globals;

/* API interfaces */
extern GLjumptable wiregl_im_api[];
extern GLapi wiregl_pack_api[];
extern GLapi vertexBBapi[];
extern GLapi vertexBBSIMDapi[];
extern GLapi vertexCapi[];
#ifdef WINDOWS
extern GLapi vertexSafeCompressBBapi[];
extern GLapi vertexSafeCompressCapi[];
extern GLapi vertexCompressBBapi[];
extern GLapi vertexCompressCapi[];
#endif


void wireGLInitGlobals( void );
void wireGLClientParseConfig( void );
void wireGLSendBufferInit( WireGLSendBuffer *send_buf, void *pack, unsigned int size );


void wireGLSwapBuffers( void );
void wireGLCreateContext( void );
void wireGLMakeCurrent( void );

void *wireGLAllocPacket( unsigned int size );
void wireGLFreePacket( void *packet );
void wireGLSendPacket( WireGLOpcode op, void *packet );
void wireGLMultiSendPacket( GLcontext *ctx, WireGLOpcode op, void *packet );
void wireGLFlushBuffer( GLcontext *ctx );
void wireGLFlushAll( GLcontext *ctx );
void wireGLFlushAllWithOpcode( GLcontext *ctx, WireGLOpcode op, unsigned int n_dat, void *data );
void wireGLClientReceive( void *buf, unsigned int len, WireGLFreeFunc freeFunc );
void wireGLPipeServerBufferAppend( WireGLPipeServer *pipe, WireGLSendBuffer *src );
void wireGLSendPipeServerBuffer( WireGLPipeServer *pipe );
void wireGLSyncWithPipes( void );
void wireGLVpinch (GLcontext *ctx);

#define GET_BUFFERED_POINTER( len ) \
  data_ptr = __wiregl_globals.buffer.data_current; \
  if (data_ptr + (len) > __wiregl_globals.buffer.data_end ) \
  { \
    wireGLFlushBuffer( __wiregl_globals.context ); \
	data_ptr = __wiregl_globals.buffer.data_current; \
    wireGLAssert( data_ptr + (len) <= __wiregl_globals.buffer.data_end ); \
  } \
  __wiregl_globals.buffer.data_current += (len)

#define GET_BUFFERED_POINTER_NO_ARGS( ) \
  GET_BUFFERED_POINTER( 4 );  \
  WRITE_DATA( 0, GLuint, 0xdeadbeef )

#define GET_BUFFERED_COUNT_POINTER( len ) \
  data_ptr = __wiregl_globals.buffer.data_current; \
  if (data_ptr + (len) > __wiregl_globals.buffer.data_end || \
	  __wiregl_globals.current.vtx_count >= __wiregl_globals.current.vtx_max) \
  { \
    wireGLFlushBuffer( __wiregl_globals.context ); \
	data_ptr = __wiregl_globals.buffer.data_current; \
  } \
  __wiregl_globals.buffer.data_current += (len); \
  __wiregl_globals.current.vtx_count++; \

#define GET_BUFFERED_COUNT_POINTER_4VEC( len ) \
  data_ptr = __wiregl_globals.buffer.data_current; \
  if (data_ptr + 16 > __wiregl_globals.buffer.data_end || \
	  __wiregl_globals.current.vtx_count >= __wiregl_globals.current.vtx_max) \
  { \
    wireGLFlushBuffer( __wiregl_globals.context ); \
	data_ptr = __wiregl_globals.buffer.data_current; \
  } \
  __wiregl_globals.buffer.data_current += (len); \
  __wiregl_globals.current.vtx_count++; \



#define WRITE_DATA( offset, type, data ) \
  *( (type *) (data_ptr + (offset))) = (data)

#ifdef WIREGL_UNALIGNED_ACCESS_OKAY
#define WRITE_DOUBLE( offset, data ) \
  WRITE_DATA( offset, GLdouble, data )
#else
#define WRITE_DOUBLE( offset, data ) \
  wireGLWriteUnalignedDouble( data_ptr + (offset), (data) )
#endif

#define WRITE_OPCODE( opcode ) \
  *(__wiregl_globals.buffer.opcode_current--) = (unsigned char) opcode

#define WRITE_NETWORK_POINTER( offset, data ) \
  wireGLNetworkPointerWrite( (WireGLNetworkPointer *) ( data_ptr + (offset) ), (data) )

#endif /* WIREGL_CLIENT_H */

