/* 
** opengl_stub/config.c 
** http://graphics.stanford.edu/software/wiregl
**
** Copyright 2000
** The Board of Trustees of The Leland Stanford Junior University.
** All rights reserved.
**
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or
** modify this software, provided that the above copyright notice and
** this permission notice appear in all copies of this software.  No
** part of this software or any derivatives thereof may be used in
** graphics systems for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind,
** express, implied or otherwise, including without limitation, any
** warranty of merchantability or fitness for a particular purpose.
*/

#include <stdio.h>
#include <stdlib.h>

#include "wiregl/include/wiregl_util.h"
#include "wiregl/include/wiregl_client.h"
#include "wiregl/include/wiregl_common.h"
#include "wiregl/include/wiregl_instrument.h"

#ifndef WINDOWS
#define WIREGL_DEFAULT_PATH "."
#else
#define WIREGL_DEFAULT_PATH  "." WIREGL_PATH_SEPARATOR "d:\\work\\wiregl\\opengl_stub" WIREGL_PATH_SEPARATOR "c:\\work\\wiregl\\opengl_stub"
#endif
#define WIREGL_DEFAULT_CONF_FILENAME "wiregl.conf"

#ifdef SIMD
	__m128 pad16byte;
#endif
WireGLGlobals __wiregl_globals;

static WireGLPipeServer *cur_pipeserver = NULL;

static void InitiateServerConnections( void )
{
	WireGLConnectionRequest *request;
	int i;
	int use_ring = __wiregl_globals.use_ring;

	wireGLNetInit( wireGLClientReceive, NULL );

	request = (WireGLConnectionRequest *) wireGLAlloc (sizeof (*request) * __wiregl_globals.num_servers);
	memset(request, 0, sizeof (request) * __wiregl_globals.num_servers);
	
	for ( i = 0 ; i < __wiregl_globals.num_servers; i++ )
	{
		WireGLPipeServer *p = __wiregl_globals.servers + i;
		WireGLConnectionRequest *q = request+i;

		q->max_send      = __wiregl_globals.pipe_buffer_size;
		q->window_x      = p->window_x;
		q->window_y      = p->window_y;
		q->window_width  = p->w;
		q->window_height = p->h;
		q->feather       = p->feather;
		q->depth_bits    = __wiregl_globals.depth_bits;
		q->stencil_bits  = __wiregl_globals.stencil_bits;

		q->use_l2        = p->use_l2;
		
		*(q->peer) = '\0';

		if (use_ring && i+1 < __wiregl_globals.num_servers) {
			if (strlen ((p+1)->server) > WIREGL_REQUEST_PEER_NAME_LENGTH)
				wireGLAbort();
			strcpy (q->peer, (p+1)->server);
		}
	}

	for ( i = 0 ; i < __wiregl_globals.num_servers; i++ )
	{
		WireGLPipeServer *p = __wiregl_globals.servers + i;
		WireGLConnectionRequest *q = request + i;
		
		p->conn = wireGLConnectToServer( p->server, WIREGL_TCPIP_DEFAULT_PORT,
										 q);

		wireGLFinishConnectToServer( p->conn );

		wireGLSendBufferInit( &p->buffer, wireGLNetAlloc( p->conn ),
							  wireGLNetMaxSend( ) );
		
		if (use_ring) break;
	}
	
	wireGLFree(request);

	/* If ring is enabled, reconfigure the globals to send the stream to 
	** only one server. */
	if (use_ring) {
		__wiregl_globals.broadcast = 1;
		__wiregl_globals.num_servers = 1;
	}
}

typedef struct WireGLParameterStruct {
	char *str;
	int value;
} WireGLParameterStruct;

static WireGLParameterStruct wiregl_boolean[] = {
	{ "0", 0 },
	{ "1", 1 },
	{ "disable", 0 },
	{ "enable", 1 },
	{ "n", 0 },
	{ "y", 1 },
	{ "no", 0 },
	{ "yes", 1 },
	{ "off", 0 },
	{ "on", 1 },
	{ NULL, 0 }
};

static int ParseParameter( char *buf , WireGLParameterStruct *parsers )
{
	WireGLParameterStruct *ps;
	for (ps = parsers ; ps->str != NULL ; ps++)
	{
		if ( !strcmp( buf, ps->str ) )
		{
			return ps->value;
		}
	}

	wireGLSimpleError( "Unrecognized parameter: %s", buf );
	return -1;
}

static void ParseNumPipeservers( char *buf )
{
	int i;

	if ( sscanf( buf, "%d", &(__wiregl_globals.num_servers) ) != 1 )
		wireGLSimpleError( "bad num pipeserver : %s\n", buf );

	__wiregl_globals.servers = 
		(WireGLPipeServer *) wireGLAlloc( __wiregl_globals.num_servers * sizeof( *__wiregl_globals.servers ));
	for ( i = 0 ; i < __wiregl_globals.num_servers; i++)
	{
		WireGLPipeServer *p = __wiregl_globals.servers + i;

		// don't initialize the SendBuffer here, because the
		// pipe_buffer_size isn't (necessarily) finalized

		p->writeback = 0;
		p->index = i;
		p->w = 1024;
		p->h = 768;
		p->mural_x = 0;
		p->mural_y = 0;
		p->window_x = 0;
		p->window_y = 0;
		p->feather.top_dice = 0;
		p->feather.top_size = 0;
		p->feather.top[0] = 0.0f;
		p->feather.top[1] = 0.0f;
		p->feather.top[2] = 0.0f;
		p->feather.top[3] = 0.0f;
		p->feather.bottom_dice = 0;
		p->feather.bottom_size = 0;
		p->feather.bottom[0] = 0.0f;
		p->feather.bottom[1] = 0.0f;
		p->feather.bottom[2] = 0.0f;
		p->feather.bottom[3] = 0.0f;
		p->feather.left_dice = 0;
		p->feather.left_size = 0;
		p->feather.left[0] = 0.0f;
		p->feather.left[1] = 0.0f;
		p->feather.left[2] = 0.0f;
		p->feather.left[3] = 0.0f;
		p->feather.right_dice = 0;
		p->feather.right_size = 0;
		p->feather.right[0] = 0.0f;
		p->feather.right[1] = 0.0f;
		p->feather.right[2] = 0.0f;
		p->feather.right[3] = 0.0f;
		memset(p->feather.overlay, 0, sizeof(p->feather.overlay) );

		p->use_l2 = 0;
	}
}

static void ParseSetPipeserver( char *buf )
{
	int p;
	if (!__wiregl_globals.num_servers)
	{
		wireGLSimpleError( "I don't know how many pipeservers "
						   "there are yet!" );
	}
	
	if ( sscanf( buf, "%d", &p ) != 1 )
		wireGLSimpleError( "bad set_pipeserver : %s\n", buf );
	if (p < 1 || p > __wiregl_globals.num_servers)
		wireGLSimpleError( "Unknown pipeserver: %d (they start from one)", p );
	cur_pipeserver = __wiregl_globals.servers + p - 1;
}

static void ParseServer( char *buf )
{
	cur_pipeserver->server = strdup( buf );
}

static void ParseMuralExtent( char *buf )
{
	if ( sscanf( buf, "%d %d %d %d", 
				 &(cur_pipeserver->mural_x), &(cur_pipeserver->mural_y),
				 &(cur_pipeserver->w),       &(cur_pipeserver->h) ) != 4 )
		wireGLSimpleError( "bad mural_extent : %s\n", buf );

	if (cur_pipeserver->mural_x + cur_pipeserver->w > __wiregl_globals.mural.width)
		__wiregl_globals.mural.width = cur_pipeserver->mural_x + cur_pipeserver->w;
	if (cur_pipeserver->mural_y + cur_pipeserver->h > __wiregl_globals.mural.height)
		__wiregl_globals.mural.height = cur_pipeserver->mural_y + cur_pipeserver->h;
}

static void ParseSwapBarrier( char *buf )
{
#if 0
	if ( __wiregl_globals.num_servers > 1 )
	{
		WireGLConnection *bconn;
		__wiregl_globals.swap_barrier = 1;
		bconn = wireGLBarrierConnect( buf );
		wireGLBarrierCreate( bconn, WIREGL_PIPESERVER_BARRIER,
							 __wiregl_globals.num_servers );
		__wiregl_globals.barrier_server = strdup( buf );
	}
#endif
	((void) buf);
}

static void ParsePrintStats( char *buf )
{
	char name[1024];
	int i;

	if (!strcmp( buf, "bytes_per_server" ))
	{
		__wiregl_globals.print.bytes_per_server = 1;
		wireGLInstrumentRegisterPerServerCounter( WIREGL_BYTES_PER_SERVER, __wiregl_globals.num_servers, "Bytes per server" );
	}
	else if (!strcmp( buf, "bytes_per_frame_per_server" ) )
	{
		__wiregl_globals.print.bytes_per_frame_per_server = 1;
		for (i = 0 ; i < __wiregl_globals.num_servers ; i++)
		{
			sprintf( name, "Bytes per frame[%d]", i+1 );
			wireGLInstrumentRegisterPerFrameCounter( WIREGL_BYTES_PER_SERVER_PER_FRAME + i, name );
		}
	}
	else if (!strcmp( buf, "min_bytes_sent" ) )
	{
		__wiregl_globals.print.min_bytes_sent = 1;
		wireGLInstrumentRegisterCounter( WIREGL_MIN_BYTES_SENT, "Minimum bytes to be sent" );
	}
	else if (!strcmp( buf, "frame_rate" ) )
	{
		__wiregl_globals.print.frame_rate = 1;
		wireGLInstrumentRegisterTimer( WIREGL_FRAME_TIMER, "Frame time" );
		wireGLInstrumentRegisterPerFrameCounter( WIREGL_FRAME_TIMER, "Frame rate" );
		wireGLInstrumentStartTimer( WIREGL_FRAME_TIMER );
	}
	else if (!strcmp( buf, "flush_time" ) )
	{
		__wiregl_globals.print.flush_time = 1;
		wireGLInstrumentRegisterTimer( WIREGL_FLUSH_TIMER, "Flush time" );
		wireGLInstrumentRegisterPerFrameCounter( WIREGL_FLUSH_TIMER, "Flush time" );
	}
	else if (!strcmp( buf, "send_time" ) )
	{
		__wiregl_globals.print.send_time = 1;
		wireGLInstrumentRegisterTimer( WIREGL_SEND_TIMER, "Send time" );
		wireGLInstrumentRegisterPerFrameCounter( WIREGL_SEND_TIMER, "Send time" );
	}
	else if (!strcmp( buf, "diff_context_time" ) )
	{
		__wiregl_globals.print.diff_context_time = 1;
		wireGLInstrumentRegisterTimer( WIREGL_DIFF_CONTEXT_TIMER, "Diff context time" );
		wireGLInstrumentRegisterPerFrameCounter( WIREGL_DIFF_CONTEXT_TIMER, "Diff context time" );
	}
	else if (!strcmp( buf, "bytes_sent" ) )
	{
		__wiregl_globals.print.total_bytes_sent = 1;
		wireGLInstrumentRegisterCounter( WIREGL_TOTAL_BYTES_SENT, "Total bytes sent" );
	}
	else if (!strcmp( buf, "total_time" ) )
	{
		__wiregl_globals.print.total_time = 1;
		wireGLInstrumentRegisterTimer( WIREGL_TOTAL_TIME, "Total time" );
		wireGLInstrumentStartTimer( WIREGL_TOTAL_TIME );
	}
	else 
    {
		wireGLSimpleError( "bad print_stats : %s\n", buf );
	}
}

static void ParseStatsFile( char *buf ) 
{
	wireGLInstrumentToFile( buf );
}

static void ParseMuralWindow( char *buf )
{
	if ( sscanf( buf, "%d %d", &(cur_pipeserver->window_x), 
				 &(cur_pipeserver->window_y) ) != 2 )
		wireGLSimpleError( "bad mural_window : %s\n", buf );
}

static void ParsePipeBufferSize( char *buf )
{
	int buffer_size;
	if ( sscanf( buf, "%d", &buffer_size ) != 1 )
		wireGLSimpleError( "bad pipe_buffer_size : %s\n", buf );
	__wiregl_globals.pipe_buffer_size = buffer_size;
}

void __ParseFeather( char *buf, int *size, float cp[])
{
	sscanf( buf, "%d %f %f %f %f", size, cp, cp+1, cp+2, cp+3);
}
		
void ParseFeatherTop( char *buf ) {
	__ParseFeather( buf, &cur_pipeserver->feather.top_size, 
		cur_pipeserver->feather.top);
}

void ParseFeatherBottom( char *buf ) {
	__ParseFeather( buf, &cur_pipeserver->feather.bottom_size,
		cur_pipeserver->feather.bottom);
}

void ParseFeatherLeft( char *buf ) {
	__ParseFeather( buf, &cur_pipeserver->feather.left_size,
		cur_pipeserver->feather.left);
}

void ParseFeatherRight( char *buf ) {
	__ParseFeather( buf, &cur_pipeserver->feather.right_size,
		cur_pipeserver->feather.right);
}

void ParseFeatherOverlay( char *buf ) {
	strncpy(cur_pipeserver->feather.overlay, buf, 256);
	cur_pipeserver->feather.overlay[255] = 0;
}

void ParseDepthBits( char *buf ) {
	if( sscanf( buf, "%d", &__wiregl_globals.depth_bits) != 1)
		wireGLSimpleError( "bad depth_bits : %s\n", buf );
}

void ParseStencilBits( char *buf ) {
	if( sscanf( buf, "%d", &__wiregl_globals.stencil_bits) != 1)
		wireGLSimpleError( "bad stencil_bits : %s\n", buf );
}

static void ParseGuiError( char *buf )
{
	__wiregl_globals.gui_error = ParseParameter( buf, wiregl_boolean );
}

static void ParseDrawBBox( char *buf )
{
	__wiregl_globals.draw_bbox = ParseParameter( buf, wiregl_boolean );
}

static void ParseWarnLevel( char *buf )
{
	__wiregl_globals.warn_level_str = strdup( buf );
}

static void ParseBucketSize( char *buf )
{
	int size;
	if ( sscanf( buf, "%d", &size ) != 1 || size < 3 )
		wireGLSimpleError( "bucket_size: bad value (%s)\n", buf );
	__wiregl_globals.bucket_size = size;
}

static void ParseBroadcast( char *buf )
{
	__wiregl_globals.broadcast = ParseParameter( buf, wiregl_boolean );
}

static void ParsePackOnly( char *buf )
{
	__wiregl_globals.pack_only = ParseParameter( buf, wiregl_boolean );
}

static void ParseResolveArrays( char *buf )
{
	__wiregl_globals.resolve_arrays = ParseParameter( buf, wiregl_boolean );
}

static void ParseStateErrorGui( char *buf )
{
	__wiregl_globals.state_error_gui = ParseParameter( buf, wiregl_boolean );
}

static void ParseSyncOnFinish( char *buf )
{
	__wiregl_globals.sync_on_finish = ParseParameter( buf, wiregl_boolean );
}

static void ParseSyncOnSwap( char *buf )
{
	__wiregl_globals.sync_on_swap = ParseParameter( buf, wiregl_boolean );
}

static void ParseSplitBeginEnd( char *buf )
{
	__wiregl_globals.split_begin_end = ParseParameter( buf, wiregl_boolean );
}

static void ParseLightning2( char *buf )
{
	cur_pipeserver->use_l2 = ParseParameter( buf, wiregl_boolean );
}

static void ParseUseRing( char *buf )
{
	__wiregl_globals.use_ring = ParseParameter( buf, wiregl_boolean );
}

static void ParseUseSimd( char *buf )
{
	__wiregl_globals.use_simd = ParseParameter( buf, wiregl_boolean );
}


static WireGLParseStruct wiregl_parse[] = {
	{ "num_pipeservers", ParseNumPipeservers },
	{ "set_pipeserver", ParseSetPipeserver },
	{ "server", ParseServer },
	{ "mural_extent", ParseMuralExtent },
	{ "swap_barrier", ParseSwapBarrier },
	{ "print_stats", ParsePrintStats },
	{ "stats_file", ParseStatsFile },
	{ "mural_window", ParseMuralWindow },
	{ "pipe_buffer_size", ParsePipeBufferSize },
	{ "gui_error", ParseGuiError },
	{ "draw_bbox", ParseDrawBBox },
	{ "warn_level", ParseWarnLevel },
	{ "bucket_size", ParseBucketSize },
	{ "broadcast", ParseBroadcast },
	{ "pack_only", ParsePackOnly },
	{ "resolve_arrays", ParseResolveArrays },
	{ "state_error_gui", ParseStateErrorGui },
    { "sync_on_finish", ParseSyncOnFinish },
    { "sync_on_swap", ParseSyncOnSwap },
	{ "feather_top", ParseFeatherTop },
	{ "feather_bottom", ParseFeatherBottom },
	{ "feather_left", ParseFeatherLeft },
	{ "feather_right", ParseFeatherRight },
	{ "feather_overlay", ParseFeatherOverlay },
	{ "depth_bits", ParseDepthBits },
	{ "stencil_bits", ParseStencilBits },
	{ "split_begin_end", ParseSplitBeginEnd },
	{ "lightning2", ParseLightning2 },
	{ "use_ring", ParseUseRing },
	{ "use_simd", ParseUseSimd },
	{ NULL, NULL }
};

void wireGLInitGlobals( void )
{
	__wiregl_globals.mural.width = 0;
	__wiregl_globals.mural.height = 0;
	__wiregl_globals.print.bytes_per_server = 0;
	__wiregl_globals.print.bytes_per_frame_per_server = 0;
	__wiregl_globals.print.min_bytes_sent = 0;
	__wiregl_globals.print.frame_rate = 0;
	__wiregl_globals.print.flush_time = 0;
	__wiregl_globals.print.send_time = 0;
	__wiregl_globals.print.diff_context_time = 0;
	__wiregl_globals.print.total_bytes_sent = 0;
	__wiregl_globals.print.total_time = 0;
	__wiregl_globals.pipe_buffer_size = 8 * 1024;
	__wiregl_globals.swap_barrier = 0;
	__wiregl_globals.barrier_server = NULL;

#ifndef WINDOWS
	__wiregl_globals.depth_bits = 8;
	__wiregl_globals.stencil_bits = 0;
#else
	__wiregl_globals.depth_bits = 32;
	__wiregl_globals.stencil_bits = 0;
#endif /* !WINDOWS */

	__wiregl_globals.normal_compression = 0;
	__wiregl_globals.vertex_compression = 0;
	__wiregl_globals.unsafe_compression = 0;
	__wiregl_globals.compression_depthbits = -1;
	__wiregl_globals.gui_error       = 0;
	__wiregl_globals.draw_bbox       = 0;
	__wiregl_globals.warn_level_str  = "critical";
	__wiregl_globals.bucket_size     = GL_MAXINT;
	__wiregl_globals.broadcast       = 0;
	__wiregl_globals.pack_only       = 0;
	__wiregl_globals.resolve_arrays  = 0;
	__wiregl_globals.state_error_gui = 0;
	__wiregl_globals.sync_on_finish  = 0;
	__wiregl_globals.sync_on_swap    = 0;
	__wiregl_globals.split_begin_end = 1;
	__wiregl_globals.use_ring        = 0;
	__wiregl_globals.use_simd        = 0;
}

void wireGLClientParseConfig( void )
{
	FILE *fp = NULL;
	char fname[1024];

	char *wiregl_conf, *wiregl_path, *temp;

	int found_it = 0;

	wiregl_conf = getenv( "WIREGL_CONF" );
	if ( !wiregl_conf ) 
	{
		wiregl_path = getenv( "WIREGL_PATH" );
		if ( !wiregl_path )
		{
			wiregl_path = strdup( WIREGL_DEFAULT_PATH );
		}

		temp = wiregl_path;
		while ( temp && *temp )
		{
			char *colon = strchr( temp, WIREGL_PATH_SEPARATOR[0] );
			if ( colon )
			{
				strncpy( fname, temp, colon-temp );
				fname[colon-temp] = '\0';
			}
			else
				strcpy( fname, temp );
			strcat( fname, "/" );
			strcat( fname, WIREGL_DEFAULT_CONF_FILENAME );
			fp = fopen(fname, "r");
			if ( fp )
			{
				fclose(fp);
				found_it = 1;
				break;
			}
			temp = colon ? ( colon + 1 ) : NULL;
		}

		if ( !found_it )
		{
			wireGLSimpleError( "wireGLClientParseConfig: Can't find %s in "
							   "WIREGL_PATH (%s)", 
							   WIREGL_DEFAULT_CONF_FILENAME, wiregl_path );
		}
		wiregl_conf = fname;
	}

	__wiregl_globals.num_servers = 0;

	wireGLParseFile( wiregl_conf, wiregl_parse );

	if ( getenv( "WIREGL_WARN_LEVEL" ) )
		__wiregl_globals.warn_level_str = 
			strdup( getenv( "WIREGL_WARN_LEVEL" ) );
	if ( getenv( "WIREGL_GUI_ERROR" ) )
		__wiregl_globals.gui_error = 1;
	if ( getenv( "WIREGL_DRAW_BBOX" ) )
		__wiregl_globals.draw_bbox = 1;
	if ( getenv( "WIREGL_BROADCAST" ) )
		__wiregl_globals.broadcast = 1;
	if ( getenv( "WIREGL_PACK_ONLY" ) )
		__wiregl_globals.pack_only = 1;
	if ( getenv( "WIREGL_RESOLVE_ARRAYS" ) )
		__wiregl_globals.resolve_arrays = 1;
	if ( getenv( "WIREGL_STATE_ERROR_GUI" ) )
		__wiregl_globals.state_error_gui = 1;

	InitiateServerConnections( );
}
